/*-
 * Copyright 2022,2023 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * iterate over a body and call callback for anything which looks like a url
 */

#include "config.h"

#include <string.h>	// memset() for INIT_OBJ()
#include <stdlib.h>	// strtoul()
#include <stdio.h>

#include <cache/cache.h>
#include <vct.h>

#include "zfr_iter.h"

/* This function exists because flexelint does not understand (p += strlen(s))
 * in place
 */

static inline int
pincr(const char **p, size_t l)
{
	*p += l;
	return (1);
}

#define tok(p, e, s) (							\
	    pdiff(p,e) >= strlen(s) &&					\
	    strncmp(p, s, strlen(s)) == 0 &&				\
	    pincr(&p, strlen(s))					\
	    )

int
zfr_iter(void *priv, unsigned flush, const void *ptr, ssize_t alen)
{
	const char *pp, *p, *e, *h, *u;
	struct zfr_iter_s *zis;
	char *hh, *uu;
	size_t l;

	CAST_OBJ_NOTNULL(zis, priv, ZFR_ITER_MAGIC);

	AZ(zis->end_seen);
	zis->end_seen = ((flush & OBJ_ITER_END) != 0);

#ifdef DEBUG
	printf(
	    "buf %zu >%.*s<\t"
	    "ptr %zu >%.*s<\n",
	    zis->len, (int)zis->len, zis->buf,
	    len, (int)len, (const char *)ptr);
#endif

	if (zis->len > 0)
		AN(zis->buf);

	if (alen == 0 && zis->len == 0)
		return (0);

	assert(alen >= 0);
	l = (size_t) alen;

	if (zis->len == 0)
		p = ptr;
	else {
		zis->buf = realloc(zis->buf, zis->len + l);
		AN(zis->buf);
		memcpy(zis->buf + zis->len, ptr, l);
		zis->len += l;

		p = zis->buf;
		l = zis->len;
	}
	AN(p);

	pp = p;
	e = p + l;
	while (p < e) {
		if (vct_islws(*p)) {
			p++;
			continue;
		}
		if (tok(p, e, "https://") ||
		    tok(p, e, "http://")) {
			p -= 2;
			// no need to keep this prefix in the buffer
			pp = p;
		}
		h = NULL;
		if (tok(p, e, "//")) {
			h = u = p;
			p -= 2;
			while (u < e && *u != '/')
				u++;
		} else
			u = p;

		while (p < e && !vct_islws(*p))
			p++;

		if (*u != '/')
			break;

		if (p < e || flush & OBJ_ITER_END) {
			/* match! */
			hh = NULL;
			if (h) {
				assert(u > h);
				hh = strndup(h, pdiff(h, u));
				AN(hh);
			}
			assert(p > u);
			uu = strndup(u, pdiff(u, p));
			AN(uu);
			AN(zis->func);
			zis->func(zis->priv, uu, hh);
			free(hh);
			free(uu);
			pp = p;
		}
	}

	assert(e >= pp);
	l = pdiff(pp, e);
	if (flush & OBJ_ITER_END || l == 0) {
		zis->len = 0;
		return (0);
	}
	if (zis->len)
		memmove(zis->buf, pp, l);
	else {
		zis->buf = realloc(zis->buf, l);
		memcpy(zis->buf, pp, l);
	}
	zis->len = l;
	return (0);
}

void
zfr_iter_fini(struct zfr_iter_s *zis)
{

	CHECK_OBJ_NOTNULL(zis, ZFR_ITER_MAGIC);

	if (zis->len > 0 && zis->end_seen == 0)
		(void) zfr_iter(zis, OBJ_ITER_END, "", (ssize_t)0);
	free(zis->buf);
	zis->buf = NULL;
	zis->len = 0;
}

#ifdef TEST_DRIVER
#include <stdio.h>

struct expect {
	const char *u, *h, **pfx;
};

const char *pfx_host[7] = {
	"https://",
	"http://",
	"//",
	"  https://",
	" http://",
	"   //",
	NULL
};

const char *pfx_nohost[2] = {
	"",
	NULL
};

static struct expect testcase[7] = {
	{ .u = "/url", .h = "host", .pfx = pfx_host },
	{ .u = "/",    .h = "host", .pfx = pfx_host },
	{ .u = "///",  .h = "ho-t", .pfx = pfx_host },
	{ .u = "/url", .h = NULL,   .pfx = pfx_nohost },
	{ .u = "/",    .h = NULL,   .pfx = pfx_nohost },
	{ .u = "/a",   .h = NULL,   .pfx = pfx_nohost },
	{ NULL }
};

struct cb_want_priv {
	unsigned		magic;
#define CBW_MAGIC		0x79d8b905
	unsigned		count;
	const struct expect	*e;
};

static void
cb_not(void *priv, const char *u, const char *h)
{
	WRONG("not to be called");
}

static void
cb_want(void *priv, const char *u, const char *h)
{
	struct cb_want_priv *p;
	const struct expect *e;

	CAST_OBJ_NOTNULL(p, priv, CBW_MAGIC);
	e = p->e;

	if (e->u == NULL)
		AZ(u);
	else
		AZ(strcmp(e->u, u));

	if (e->h == NULL)
		AZ(h);
	else
		AZ(strcmp(e->h, h));

	p->count++;
}

static void
t_steps(const struct expect *want, unsigned n, const char *s)
{
	struct cb_want_priv wp[1];
	struct zfr_iter_s zis[1];
	size_t l, step;
	const char *t;

	INIT_OBJ(zis, ZFR_ITER_MAGIC);
	INIT_OBJ(wp, CBW_MAGIC);

	wp->e = want;
	zis->priv = wp;
	zis->func = cb_want;
	for (step = 1; step < strlen(s); step++) {
		wp->count = 0;
		zis->end_seen = 0;
		for (t = s, l = strlen(t);
		     l > step;
		     l -= step, t += step)
			zfr_iter(zis, 0, t, step);
		AN(l);
		zfr_iter(zis, OBJ_ITER_END, t, l);
		AN(zis->end_seen);
		assert(wp->count == n);
	}
}

int
main(void) {
	const struct expect *want;
	struct zfr_iter_s zis[1];
	const char **p, *t;
	char s[256], u[64];
	unsigned n;
	size_t l;

	// base test of tok()
	strcpy(s, "https://");
	t = s;
	assert(tok(t, strchr(s, '\0'), "https://"));


	for (want = testcase; want->u != NULL; want++) {
		INIT_OBJ(zis, ZFR_ITER_MAGIC);
		zis->priv = NULL;
		zis->func = cb_not;
		zfr_iter(zis, 0, "", 0);

		n = 0;
		*s = '\0';
		for (p = want->pfx; *p != NULL; p++) {
			AN(p);
			AN(*p);
			if (**p == '\0')
				l = snprintf(u, sizeof u, "\t%s", want->u);
			else {
				l = snprintf(u, sizeof u, "\t%s%s%s",
				    *p, want->h, want->u);
			}
			assert(l < sizeof u);

			if (sizeof s - strlen(s) - 1 < l)
				break;
			(void) strcat(s, u);
			n++;

			t_steps(want, n, s);
		}
	}

	zfr_iter_fini(zis);
	printf("OK\n");

	return (0);
}
#endif
